import { h as hash, c as calcDuration, S as Store, C as CalculationHelper, s as simpleShallowEquals, d as defaultEquals, m as makeSelector } from "./store.js";
import { e, a, f, r, g, b } from "./store.js";
class Cache {
  constructor(factory, cacheTime) {
    this.factory = factory;
    this.cacheTime = cacheTime;
    this.cache = /* @__PURE__ */ new Map();
    this.interval = this.cacheTime ? setInterval(() => this.cleanup(), Math.max(this.cacheTime / 10, 1)) : void 0;
  }
  cleanup() {
    var _a;
    const cutoff = this.now() - (this.cacheTime ?? 0);
    for (const [key, entry] of [...this.cache.entries()]) {
      if (entry.ref && entry.t <= cutoff) {
        delete entry.ref;
      }
      if (!entry.ref && !((_a = entry.weakRef) == null ? void 0 : _a.deref())) {
        this.cache.delete(key);
      }
    }
  }
  get(...args) {
    var _a;
    const key = hash(args);
    let entry = this.cache.get(key);
    let value = (entry == null ? void 0 : entry.ref) ?? ((_a = entry == null ? void 0 : entry.weakRef) == null ? void 0 : _a.deref());
    if (!entry || !value) {
      value = this.factory(...args);
      entry = {
        t: this.now(),
        ref: value,
        weakRef: typeof WeakRef !== "undefined" ? new WeakRef(value) : void 0
      };
      this.cache.set(key, entry);
    } else {
      entry.t = this.now();
      entry.ref ?? value;
    }
    return value;
  }
  values() {
    return [...this.cache.values()].map((entry) => {
      var _a;
      return entry.ref ?? ((_a = entry.weakRef) == null ? void 0 : _a.deref());
    }).filter((value) => !!value);
  }
  stop() {
    if (this.interval) {
      clearInterval(this.interval);
    }
  }
  stats() {
    return {
      count: this.cache.size,
      withRef: [...this.cache.values()].filter((x) => !!x.ref).length,
      withWeakRef: [...this.cache.values()].filter((x) => {
        var _a;
        return !!((_a = x.weakRef) == null ? void 0 : _a.deref());
      }).length
    };
  }
  now() {
    return performance.now();
  }
}
class ResourceGroup extends Set {
  invalidate() {
    for (const resource of this) {
      resource.invalidate();
    }
  }
  clear() {
    for (const resource of this) {
      resource.clear();
    }
  }
}
const allResources = new ResourceGroup();
const fetchStoreStateEquals = (equals = defaultEquals) => (a2, b2) => {
  const { value: av, ...ar } = a2;
  const { value: bv, ...br } = b2;
  return simpleShallowEquals(ar, br) && (ar.status !== "value" || equals(av, bv));
};
const createRef = () => Math.random().toString(36).slice(2);
class FetchStore extends Store {
  constructor(fetchFn, options = {}) {
    super({
      status: "pending",
      isStale: true,
      isUpdating: false,
      ref: createRef()
    });
    this.fetchFn = fetchFn;
    this.options = options;
    this.calculationHelper = new CalculationHelper({
      calculate: ({ use, useFetch }) => {
        const promise = this.fetchFn.apply({ use, useFetch });
        this.setPromise(promise);
      },
      addEffect: this.addEffect,
      getValue: () => this.value.value,
      setValue: this.setValue,
      setError: this.setError,
      onInvalidate: this.invalidate
    });
  }
  update(value) {
    this.calculationHelper.stop();
    super.update(value);
  }
  async fetch(options) {
    this.calculationHelper.check();
    const { cache = "updateWhenStale" } = options ?? {};
    const { status, value, error, update, isStale } = this.value;
    if ((status === "pending" || isStale) && !update || cache === "forceUpdate") {
      this.calculationHelper.execute();
      if (status === "pending" || cache !== "backgroundUpdate") {
        return this.value.update;
      }
    }
    if (status === "value") {
      return value;
    }
    if (status === "error") {
      throw error;
    }
    return update;
  }
  setValue(value) {
    if (value instanceof Promise) {
      this.calculationHelper.stop();
      this.setPromise(value);
    } else {
      this.update({
        status: "value",
        value,
        isStale: false,
        isUpdating: false,
        ref: createRef()
      });
    }
  }
  setPromise(promise) {
    const ref = createRef();
    super.update({
      ...this.value,
      isUpdating: true,
      update: promise,
      ref
    });
    promise.then((value) => {
      if (promise === this.value.update) {
        super.update({
          status: "value",
          value,
          isStale: false,
          isUpdating: false,
          ref
        });
      }
    }).catch((error) => {
      if (promise === this.value.update) {
        super.update({
          status: "error",
          error,
          isStale: false,
          isUpdating: false,
          ref
        });
      }
    });
  }
  setError(error) {
    this.update({
      status: "error",
      error,
      isStale: false,
      isUpdating: false,
      ref: createRef()
    });
  }
  invalidate() {
    this.update({
      ...this.value,
      isStale: true,
      isUpdating: false,
      update: void 0
    });
    if (this.isActive) {
      this.calculationHelper.execute();
    }
  }
  clear() {
    this.update({
      status: "pending",
      isStale: true,
      isUpdating: false,
      ref: createRef()
    });
    if (this.isActive) {
      this.calculationHelper.execute();
    }
  }
  sub(listener, options) {
    return super.sub(listener, {
      ...options,
      equals: fetchStoreStateEquals(options == null ? void 0 : options.equals)
    });
  }
  mapValue(_selector) {
    const selector = makeSelector(_selector);
    const that = this;
    return new FetchStore(async function() {
      const value = await this.useFetch(that);
      return selector(value);
    });
  }
}
const defaultOptions = {};
function create(fetch, options) {
  return withArgs(fetch, options)();
}
function withArgs(fetch, options) {
  const { clearUnusedAfter = defaultOptions.clearUnusedAfter ?? 0, resourceGroup } = options ?? {};
  const cache = new Cache(
    (...args) => new FetchStore(function() {
      return fetch.apply(this, args);
    }, options),
    calcDuration(clearUnusedAfter)
  );
  const get = (...args) => {
    return cache.get(...args);
  };
  const invalidate = () => {
    for (const instance of cache.values()) {
      instance.invalidate();
    }
  };
  const clear = () => {
    for (const instance of cache.values()) {
      instance.clear();
    }
  };
  const resource = { invalidate, clear };
  const groups = Array.isArray(resourceGroup) ? resourceGroup : resourceGroup ? [resourceGroup] : [];
  for (const group of groups.concat(allResources)) {
    group.add(resource);
  }
  return Object.assign(get, resource);
}
const fetchStore = Object.assign(create, {
  withArgs,
  defaultOptions
});
function once(subscribe, condition) {
  return new Promise((resolve) => {
    let stopped = false;
    const cancel = subscribe(
      (value) => {
        if (stopped || condition && !condition(value)) {
          return;
        }
        resolve(value);
        stopped = true;
        setTimeout(() => cancel());
      },
      {
        runNow: !!condition
      }
    );
  });
}
export {
  ResourceGroup,
  allResources,
  e as arrayActions,
  a as derivedStore,
  fetchStore,
  f as mapActions,
  once,
  r as recordActions,
  g as setActions,
  b as store
};
//# sourceMappingURL=index.js.map
