import { CalculationHelper } from '@lib/calculationHelper';
import type { Path, Value } from '@lib/propAccess';
import { arrayActions, mapActions, recordActions, setActions } from '@lib/storeActions';
import type { Cancel, Duration, Effect, Listener, Selector, SubscribeOptions, Update, Use, UseOptions } from './commonTypes';
export type StoreActions = Record<string, (...args: any[]) => any>;
export type BoundStoreActions<T, Actions extends StoreActions> = Actions & ThisType<Store<T> & Actions>;
export interface StoreOptions {
    retain?: number;
}
export interface StoreOptionsWithActions<T, Actions extends StoreActions> extends StoreOptions {
    methods?: Actions & ThisType<Store<T> & Actions>;
}
type StoreWithActions<T, Actions extends StoreActions> = Store<T> & (Record<string, never> extends Actions ? T extends Map<any, any> ? typeof mapActions : T extends Set<any> ? typeof setActions : T extends Array<any> ? typeof arrayActions : T extends Record<any, any> ? typeof recordActions : Record<string, never> : Omit<BoundStoreActions<T, Actions>, keyof Store<T>>);
export declare class Store<T> {
    protected readonly initialValue: T;
    protected readonly options: StoreOptions;
    protected value: T;
    protected listeners: Set<Listener<void>>;
    protected effects: Map<Effect, {
        handle?: Cancel | undefined;
        retain?: number | undefined;
        timeout?: number | undefined;
    }>;
    protected notifyId: {};
    constructor(initialValue: T, options?: StoreOptions);
    get(): T;
    update(update: Update<T>): void;
    sub(listener: Listener<T>, options?: SubscribeOptions): Cancel;
    map<S>(selector: Selector<T, S>, options?: UseOptions): DerivedStore<S>;
    map<P extends Path<T>>(selector: P, options?: UseOptions): DerivedStore<Value<T, P>>;
    /** Add an effect that will be executed when the store becomes active, which means when it has at least one subscriber.
     * @param effect
     * If there is already a subscriber, the effect will be executed immediately.
     * Otherweise it will be executed as soon as the first subscription is created.
     * Every time all subscriptions are removed and the first is created again, the effect will be executed again.
     * @param retain
     * If provided, delay tearing down effects when the last subscriber is removed. This is useful if a short gap in subscriber coverage is supposed to be ignored. E.g. when switching pages, the old page might unsubscribe, while the new page subscribes immediately after.
     * @returns
     * The effect can return a teardown callback, which will be executed when the last subscription is removed and potentially the ratain time has passed.
     */
    addEffect(effect: Effect, retain?: Duration): () => void;
    /** Return whether the store is currently active, which means whether it has at least one subscriber. */
    get isActive(): boolean;
    protected onSubscribe(): void;
    protected onUnsubscribe(): void;
    protected notify(): void;
}
declare function _store<T, Actions extends StoreActions = Record<string, never>>(initialState: T, options?: StoreOptionsWithActions<T, Actions>): StoreWithActions<T, Actions>;
export declare const store: typeof _store & {
    defaultOptions: StoreOptions;
};
export declare class DerivedStore<T> extends Store<T> {
    protected calculate: (this: {
        use: Use;
    }, fns: {
        use: Use;
    }) => T;
    protected derivedFrom?: {
        store: Store<any>;
        selectors: (Selector<any, any> | string)[];
    } | undefined;
    calculationHelper: CalculationHelper<T>;
    protected valid: boolean;
    protected check?: () => void;
    protected cancel?: Cancel;
    constructor(calculate: (this: {
        use: Use;
    }, fns: {
        use: Use;
    }) => T, derivedFrom?: {
        store: Store<any>;
        selectors: (Selector<any, any> | string)[];
    } | undefined);
    get(): T;
    update(update: Update<T>): void;
    map<S>(selector: Selector<T, S>): DerivedStore<S>;
    map<P extends Path<T>>(selector: P): DerivedStore<Value<T, P>>;
    protected invalidate(): void;
}
declare function _derivedStore<T>(calculate: (this: {
    use: Use;
}, fns: {
    use: Use;
}) => T): DerivedStore<T>;
export declare const derivedStore: typeof _derivedStore;
export {};
